@@ -32,7 +32,7 @@ sub scan {
     my ($self) = @_;
     debug(2, 'assuming Apache layout "'.$self->layout->name.'"');
     $self->set(qw(core postrenew), $self->layout->restart_command)
-        unless $self->cfg->is_set(qw(core postrenew));
+        unless $self->is_set(qw(core postrenew));
     return $self->examine_http_config($self->layout->config_file);
 }
 
